import pytest
from django.db.models import Q

from data_browser.orm_debug import _format_value


@pytest.mark.parametrize(
    "value,expected",
    [
        (Q("bob"), "Q('bob')"),
        (Q(bob="fred"), "Q(bob='fred')"),
        (Q("tom", bob="fred"), "Q('tom', bob='fred')"),
        (Q("bob") & Q("fred"), "Q('bob', 'fred')"),
        (Q("bob") | Q("fred"), "Q('bob') | Q('fred')"),
        (~(Q("bob") & Q("fred")), "~Q('bob', 'fred')"),
        (~(Q("bob") | Q("fred")), "~Q(Q('bob') | Q('fred'))"),
        (~Q(bob=None), "~Q(bob=None)"),
        (~Q(~Q("bob")), "~Q(~Q('bob'))"),
    ],
)
def test_format_value(value, expected):
    assert _format_value(value) == expected



import json
import re
import uuid
from functools import lru_cache

import dateutil.parser
from dateutil import relativedelta
from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
from django.utils import dateparse, html, timezone

from .common import all_subclasses, debug_log, get_optimal_decimal_places

ASC, DSC = "asc", "dsc"


class TypeMeta(type):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.raw_type and self.element_type:
            assert self.raw_type.element_type == self.element_type.raw_type

    def __repr__(cls):
        return cls.__name__

    @property
    def default_lookup(cls):
        lookups = cls.lookups
        return list(lookups)[0] if lookups else None

    @property
    def lookups(cls):
        res = {}
        for name, type_ in cls._lookups().items():
            assert isinstance(type_, tuple), type_
            if len(type_) == 3:
                type_, keep_choices, pretty_name = type_
                assert keep_choices is not None, (cls, name)
            else:
                type_, keep_choices = type_
                pretty_name = name.replace("_", " ")
            res[name] = pretty_name, type_.name, keep_choices
        return res

    @property
    def pretty_name(cls):
        name = cls.__name__
        assert name.endswith("Type")
        return name[: -len("Type")]

    @property
    def name(cls):
        return cls.pretty_name.lower()


class BaseType(metaclass=TypeMeta):
    default_value = None
    choices = ()
    raw_type = None
    element_type = None

    def __init__(self):
        assert False

    @classmethod
    def _lookups(cls):
        return {
            "equals": (cls, True),
            "not_equals": (cls, True),
            "is_null": (IsNullType, False),
        }

    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: value

    @classmethod
    def get_formatter(cls, choices):
        return cls._get_formatter(choices)

    @staticmethod
    def _parse(value, choices):
        assert not choices
        return value

    @classmethod
    def parse(cls, value, choices):
        try:
            return cls._parse(value, choices), None
        except Exception as e:
            debug_log("Error parsing filter value", e)
            msg = str(e) if str(e) else repr(e)
            return None, msg[0].upper() + msg[1:]

    @classmethod
    def parse_lookup(cls, lookup, value, choices):
        lookups = cls.lookups
        if lookup not in lookups:
            return None, "Bad lookup"
        else:
            pretty_name, type_name, keep_choices = lookups[lookup]
            if not keep_choices:
                choices = ()
            type_ = TYPES[type_name]
            return type_.parse(value, choices)

    @classmethod
    def format_lookup(cls, lookup, value, choices):
        pretty_name, type_name, keep_choices = cls.lookups[lookup]
        if not keep_choices:
            choices = ()
        return TYPES[type_name]._get_formatter(choices)(value)

    @staticmethod
    def get_format_hints(name, data):
        return {}


class StringType(BaseType):
    default_value = ""

    @classmethod
    def _lookups(cls):
        return {
            "equals": (cls, True),
            "contains": (cls, True),
            "starts_with": (cls, True),
            "ends_with": (cls, True),
            "regex": (RegexType, False),
            "not_equals": (cls, True),
            "not_contains": (cls, True),
            "not_starts_with": (cls, True),
            "not_ends_with": (cls, True),
            "not_regex": (RegexType, False),
            "is_null": (IsNullType, False),
        }


class StringableType(BaseType):
    default_value = ""

    @staticmethod
    def _get_formatter(choices):  # pragma: no cover
        assert not choices
        return lambda value: None if value is None else str(value)


class SequenceTypeMixin:
    @classmethod
    def _lookups(cls):
        return {
            "equals": (cls, True),
            "not_equals": (cls, True),
            "gt": (cls, True, ">"),
            "gte": (cls, True, ">="),
            "lt": (cls, True, "<"),
            "lte": (cls, True, "<="),
            "is_null": (IsNullType, False),
        }


class NumberType(SequenceTypeMixin, BaseType):
    default_value = 0

    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: None if value is None else float(value)

    @staticmethod
    def _parse(value, choices):
        assert not choices
        return float(value)

    @staticmethod
    def get_format_hints(name, data):
        nums = [
            row[name] for row in data if row and row[name] and abs(row[name] > 0.0001)
        ]
        dp = get_optimal_decimal_places(nums)
        return {
            "minimumFractionDigits": dp,
            "maximumFractionDigits": dp,
            "significantFigures": 3,
            "lowCutOff": 0.0001,
            "highCutOff": 1e10,
        }


class RegexType(BaseType):
    default_value = ".*"

    @staticmethod
    @lru_cache(maxsize=None)
    def _parse(value, choices):
        assert not choices
        from django.contrib.contenttypes.models import ContentType
        from django.db.transaction import atomic

        # this is dirty
        # we need to check if the regex is going to cause a db exception
        # and not kill any in progress transaction as we check
        try:
            with atomic():
                list(ContentType.objects.filter(model__regex=value))
        except Exception as e:
            debug_log("Error parsing filter value", e)
            raise Exception("Invalid regex")
        return value


class DurationType(SequenceTypeMixin, BaseType):
    default_value = ""

    @staticmethod
    def _parse(value, choices):
        assert not choices

        if value.count(":") == 1:
            value += ":0"

        res = dateparse.parse_duration(value)
        if res is None:
            raise Exception("Duration value should be 'DD HH:MM:SS'")
        return res

    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: None if value is None else str(value)


class DateTimeParseMixin:
    _clause = re.compile(r"(\w{3,})([-+=])(\d+) *")

    _clause_types = {
        "yea": "year",
        "mont": "month",
        "day": "day",
        "hou": "hour",
        "min": "minute",
        "sec": "second",
        "mic": "microsecond",
        "wee": "week",
        "mond": relativedelta.MO,
        "tue": relativedelta.TU,
        "wed": relativedelta.WE,
        "thu": relativedelta.TH,
        "fri": relativedelta.FR,
        "sat": relativedelta.SA,
        "sun": relativedelta.SU,
    }

    for c1 in _clause_types:
        for c2 in _clause_types:
            assert c1 == c2 or not c1.startswith(c2), c2

    @staticmethod
    def _unambiguous_date_parse(value, **kwargs):
        try:
            res = {
                dateutil.parser.parse(value, dayfirst=False, yearfirst=False, **kwargs),
                dateutil.parser.parse(value, dayfirst=True, yearfirst=False, **kwargs),
                dateutil.parser.parse(value, dayfirst=False, yearfirst=True, **kwargs),
                dateutil.parser.parse(value, dayfirst=True, yearfirst=True, **kwargs),
            }
            if len(res) != 1:
                raise Exception("Ambiguous value")
            return res.pop()
        except dateutil.parser.ParserError:
            return None

    @staticmethod
    def _truncate_dt(value):
        return value.replace(hour=0, minute=0, second=0, microsecond=0)

    @classmethod
    def _parse(cls, value, choices):
        assert not choices

        d8 = r"(\d{8})"
        d422 = r"(\d{4}[^\d]+\d{1,2}[^\d]+\d{1,2}([^\d]|$))"

        res = timezone.now()

        # iterate the clauses in the expression
        for clause_str in value.split():
            clause_str = clause_str.lower()

            if clause_str == "now":
                res = timezone.now()
            elif clause_str == "today":
                res = cls._truncate_dt(timezone.now())
            elif re.match(rf"[^\d]*({d8}|{d422})", clause_str):
                # looks like some kinda iso date, roll with the defaults
                # includes T delimited like 2018-03-20T22:31:23
                res = dateutil.parser.parse(clause_str, default=cls._truncate_dt(res))
            elif cls._unambiguous_date_parse(clause_str, default=cls._truncate_dt(res)):
                # TODO could walrus the duplicate call once we drop support for py3.7
                res = cls._unambiguous_date_parse(
                    clause_str, default=cls._truncate_dt(res)
                )
            else:
                # failing that must be relative delta stuff
                match = cls._clause.fullmatch(clause_str)
                if not match:
                    raise Exception(f"Unrecognized clause '{clause_str}'")

                # cut up the clause
                field, op, val = match.groups()
                val = int(val)
                human_op = {"+": "add", "-": "subtract", "=": "set"}[op]

                # find the field
                for prefix, arg in cls._clause_types.items():
                    if field.startswith(prefix):
                        break
                else:
                    raise Exception(f"Unrecognized field '{field}'")

                # year month day etc
                if isinstance(arg, str):
                    if op == "+":
                        kwargs = {f"{arg}s": val}
                    elif op == "-":
                        kwargs = {f"{arg}s": -val}
                    else:  # op == "="
                        if arg in ["week"]:
                            raise Exception(f"'{op}' not supported for '{field}'")
                        elif arg in ["year", "month", "day"]:
                            if val <= 0:
                                raise Exception(
                                    f"Can't {human_op} '{field}' to '{val}'"
                                )
                        kwargs = {arg: val}

                # weekdays
                else:
                    if op == "=":
                        raise Exception(f"'{op}' not supported for '{field}'")
                    else:
                        if val <= 0:
                            raise Exception(f"Can't {human_op} '{val}' '{field}'s")
                        if op == "+":
                            kwargs = {"weekday": arg(val)}
                        else:  # op == "-"
                            kwargs = {"weekday": arg(-val)}

                res += relativedelta.relativedelta(**kwargs)

        return res


class DateTimeType(DateTimeParseMixin, SequenceTypeMixin, BaseType):
    default_value = "now"

    @staticmethod
    def _get_formatter(choices):
        assert not choices
        if settings.USE_TZ:
            return (
                lambda value: None
                if value is None
                else str(timezone.make_naive(value).replace(microsecond=0))
            )
        else:
            return (
                lambda value: None
                if value is None
                else str(value.replace(microsecond=0))
            )


class DateType(DateTimeParseMixin, SequenceTypeMixin, BaseType):
    default_value = "today"

    @classmethod
    def _parse(cls, value, choices):
        return super()._parse(value, choices).date()

    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: None if value is None else str(value)


class HTMLType(StringType):
    @staticmethod
    def _lookups():
        return {"is_null": (IsNullType, False)}

    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: None if value is None else html.conditional_escape(value)


class URLType(StringType):
    pass


class BooleanType(BaseType):
    default_value = True

    @staticmethod
    def _parse(value, choices):
        assert not choices

        value = value.lower()
        if value == "true":
            return True
        elif value == "false":
            return False
        else:
            raise ValueError("Expected 'true' or 'false'")

    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: None if value is None else bool(value)


class UUIDType(BaseType):
    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: None if value is None else str(value)

    @staticmethod
    def _parse(value, choices):
        assert not choices

        return uuid.UUID(value)


class UnknownType(BaseType):
    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: None if value is None else str(value)

    @staticmethod
    def _lookups():
        return {"is_null": (IsNullType, False)}


def _json_loads(value):
    try:
        return json.loads(value.strip())
    except json.JSONDecodeError:
        raise ValueError("Invalid JSON value")


class JSONFieldType(BaseType):
    default_value = "|"

    @staticmethod
    def _parse(value, choices):
        assert not choices

        value = value.strip()
        if "|" not in value:
            raise ValueError("Missing seperator '|'")
        field, value = value.split("|", 1)
        if not field:
            raise ValueError("Invalid field name")
        return [field, _json_loads(value)]


class JSONType(BaseType):
    @classmethod
    def _lookups(cls):
        return {
            "equals": (cls, True),
            "has_key": (StringType, None),
            "field_equals": (JSONFieldType, None),
            "not_equals": (cls, True),
            "not_has_key": (StringType, None),
            "not_field_equals": (JSONFieldType, None),
            "is_null": (IsNullType, False),
        }

    @staticmethod
    def _parse(value, choices):
        assert not choices

        return _json_loads(value)


class ChoiceTypeMixin:
    default_value = None

    @classmethod
    def _get_formatter(cls, choices):
        assert choices
        choices = dict(choices)
        return lambda value: choices.get(value, value)

    @classmethod
    def _parse(cls, value, choices):
        assert choices
        choices = {v: k for k, v in choices}
        if value not in choices:
            raise ValueError(f"Unknown choice '{value}'")
        return choices[value]


class StringChoiceType(ChoiceTypeMixin, BaseType):
    raw_type = StringType


class NumberChoiceType(ChoiceTypeMixin, BaseType):
    raw_type = NumberType


class IsNullType(ChoiceTypeMixin, BaseType):
    choices = [(None, "IsNull"), (True, "IsNull"), (False, "NotNull")]
    default_value = choices[0][1]

    @staticmethod
    def _lookups():
        return {"equals": (IsNullType, False)}

    @classmethod
    def _get_formatter(cls, choices):
        assert not choices
        return super()._get_formatter(cls.choices)

    @classmethod
    def _parse(cls, value, choices):
        assert not choices
        return super()._parse(value, cls.choices)


class ArrayTypeMixin:
    default_value = "[]"

    @classmethod
    def _get_formatter(cls, choices):
        element_formatter = cls.element_type._get_formatter(choices)
        return (
            lambda value: None
            if value is None
            else json.dumps(
                [element_formatter(v) for v in value], cls=DjangoJSONEncoder
            )
        )

    @classmethod
    def _lookups(cls):
        return {
            "equals": (cls, True),
            "contains": (cls.element_type, True),
            "length": (NumberType, False),
            "not_equals": (cls, True),
            "not_contains": (cls.element_type, True),
            "not_length": (NumberType, False),
            "is_null": (IsNullType, False),
        }

    @classmethod
    def _parse(cls, value, choices):
        value = _json_loads(value)
        if not isinstance(value, list):
            raise ValueError("Expected a list")
        return [cls.element_type._parse(v, choices) for v in value]


# Array types
for _cls in [
    BooleanType,
    DurationType,
    BooleanType,
    DateTimeType,
    DateType,
    UUIDType,
    URLType,
    StringType,
    NumberType,
    StringableType,
]:
    _name = f"{_cls.pretty_name}ArrayType"
    globals()[_name] = type(_name, (ArrayTypeMixin, BaseType), {"element_type": _cls})


class StringChoiceArrayType(ArrayTypeMixin, BaseType):
    element_type = StringChoiceType
    raw_type = StringArrayType  # noqa: F821


class NumberChoiceArrayType(ArrayTypeMixin, BaseType):
    element_type = NumberChoiceType
    raw_type = NumberArrayType  # noqa: F821


TYPES = {cls.name: cls for cls in all_subclasses(BaseType)}



def head_cell(field):
    return [" ".join(field.pretty_path)]


def data_cell(value, span=1):
    return [value] + [""] * (span - 1)


def v_table_head_row(fields):
    res = []
    for field in fields:
        res += head_cell(field)
    return res


def v_table_body_row(fields, row):
    res = []
    for field in fields:
        if row:
            res += data_cell(row[field.path_str])
        else:
            res += [""]
    return res


def h_table_row(field, data, span):
    res = head_cell(field)
    for col in data:
        res += data_cell(col[field.path_str], span)
    return res


def get_csv_rows(bound_query, results):
    col_fields = bound_query.col_fields
    top_title_space = len(bound_query.row_fields) - 1
    side_title_space = (
        1 - len(bound_query.row_fields) if len(bound_query.col_fields) else 0
    )
    has_body = bound_query.row_fields or bound_query.body_fields

    # col headers and data aka pivots
    for field in col_fields:
        yield [""] * top_title_space + h_table_row(
            field=field, data=results["cols"], span=len(bound_query.body_fields)
        )

    if has_body:
        # body/aggregate headers
        row = [""] * side_title_space
        row += v_table_head_row(bound_query.row_fields)
        for _ in results["cols"] or [None]:
            row += v_table_head_row(bound_query.body_fields)
        yield row

        # row headers and body
        for row_index, row_data in enumerate(results["rows"]):
            row = [""] * side_title_space
            row += v_table_body_row(bound_query.row_fields, row_data)
            for table in results["body"]:
                row += v_table_body_row(bound_query.body_fields, table[row_index])
            yield row



import contextlib
from textwrap import dedent, indent

from django.db.models import ExpressionWrapper, F, Field, Q

SPACES = "    "


def _format_value(value):
    if isinstance(value, Q):

        def format_child(child):
            if isinstance(child, tuple):
                return f"{child[0]}={_format_value(child[1])}"
            else:
                return _format_value(child)

        if len(value.children) == 1 and isinstance(value.children[0], Q):
            child = _format_value(value.children[0])
            return f"~Q({child})" if value.negated else child

        if value.connector == Q.AND:
            children = ", ".join(format_child(c) for c in value.children)
            return f"~Q({children})" if value.negated else f"Q({children})"
        else:
            children = " | ".join(f"Q({format_child(c)})" for c in value.children)
            return f"~({children})" if value.negated else children
    elif isinstance(value, Field):
        return f"{value.__class__.__name__}()"
    elif isinstance(value, ExpressionWrapper):
        expression = _format_value(value.expression)
        output_field = _format_value(value.output_field)
        return dedent(
            "ExpressionWrapper(\n"
            f"{indent(expression, SPACES)},\n"
            f"{SPACES}output_field={output_field},\n)"
        )
    else:
        return repr(value)


@contextlib.contextmanager
def _better_reprs():
    F_orig = F.__repr__
    F.__repr__ = lambda s: f"F({s.name!r})"

    yield

    F.__repr__ = F_orig


class DebugQS:
    def __init__(self, s):
        self.s = s

    def __getattr__(self, name):
        return DebugQS(f"{self.s}.{name}")

    def __call__(self, *args, **kwargs):
        if args or kwargs:
            with _better_reprs():
                flat_args = ",\n".join(
                    [_format_value(a) for a in args]
                    + [f"{n}={_format_value(v)}" for n, v in kwargs.items()]
                )
            return DebugQS(f"{self.s}(\n{indent(flat_args, SPACES)},\n)")
        else:
            return DebugQS(f"{self.s}()")

    def __getitem__(self, index):
        if isinstance(index, slice):
            assert index.start is None
            assert index.step is None
            return DebugQS(f"{self.s}[: {index.stop}]")
        else:  # pragma: no cover
            return DebugQS(f"{self.s}[index]")

    def __str__(self):
        return self.s

    def __repr__(self):
        return self.s



import re
from datetime import date, datetime, timedelta
from uuid import UUID

import pytest
import time_machine
from django.utils import timezone

from data_browser import orm_fields
from data_browser.orm_admin import OrmModel, get_fields_for_type
from data_browser.query import BoundQuery, Query, QueryField, QueryFilter
from data_browser.types import (
    ASC,
    DSC,
    BooleanType,
    DateTimeType,
    DateType,
    DurationType,
    HTMLType,
    IsNullType,
    NumberArrayType,
    NumberChoiceArrayType,
    NumberChoiceType,
    NumberType,
    RegexType,
    StringArrayType,
    StringChoiceArrayType,
    StringChoiceType,
    StringType,
    UnknownType,
    UUIDType,
)


@pytest.fixture
def query():
    return Query(
        "app.model",
        [
            QueryField("fa", False, ASC, 1),
            QueryField("fd", False, DSC, 0),
            QueryField("fn"),
        ],
        [QueryFilter("bob", "equals", "fred")],
    )


@pytest.fixture
def orm_models():
    return {
        "string": OrmModel(get_fields_for_type(StringType)),
        "number": OrmModel(get_fields_for_type(NumberType)),
        "app.model": OrmModel(
            fields={
                "fa": orm_fields.OrmConcreteField(
                    model_name="app.model",
                    name="fa",
                    pretty_name="fa",
                    type_=StringType,
                    rel_name=StringType.name,
                    choices=None,
                ),
                "fd": orm_fields.OrmConcreteField(
                    model_name="app.model",
                    name="fd",
                    pretty_name="fd",
                    type_=StringType,
                    rel_name=StringType.name,
                    choices=None,
                ),
                "fn": orm_fields.OrmCalculatedField(
                    model_name="app.model", name="fn", pretty_name="fn", func=None
                ),
                "bob": orm_fields.OrmConcreteField(
                    model_name="app.model",
                    name="bob",
                    pretty_name="bob",
                    type_=StringType,
                    rel_name=StringType.name,
                    choices=None,
                ),
                "num": orm_fields.OrmConcreteField(
                    model_name="app.model",
                    name="num",
                    pretty_name="num",
                    type_=NumberType,
                    rel_name=NumberType.name,
                    choices=None,
                ),
                "tom": orm_fields.OrmFkField(
                    model_name="app.model",
                    name="tom",
                    pretty_name="tom",
                    rel_name="app.Tom",
                    to_many=False,
                ),
            },
            admin=True,
        ),
        "app.Tom": OrmModel(
            fields={
                "jones": orm_fields.OrmConcreteField(
                    model_name="app.Tom",
                    name="jones",
                    pretty_name="jones",
                    type_=StringType,
                    rel_name=StringType.name,
                    choices=None,
                ),
                "michael": orm_fields.OrmFkField(
                    model_name="app.Tom",
                    name="michael",
                    pretty_name="michael",
                    rel_name="app.Michael",
                    to_many=False,
                ),
            },
            admin=True,
        ),
        "app.Michael": OrmModel(
            fields={
                "bolton": orm_fields.OrmConcreteField(
                    model_name="app.Michael",
                    name="bolton",
                    pretty_name="bolton",
                    type_=StringType,
                    rel_name=StringType.name,
                    choices=None,
                )
            },
            admin=True,
        ),
    }


@pytest.fixture
def bound_query(query, orm_models):
    return BoundQuery.bind(query, orm_models)


class ParseHelpers:
    choices = ()

    def check_parse(self, value, expected, error_re):
        res, err = self.type_.parse(value, self.choices)
        if expected is None:
            assert res is None
            assert re.fullmatch(error_re, err)
        else:
            assert err is None
            assert res == expected

    def check_parse_lookup(self, lookup, value, expected, error_re):
        res, err = self.type_.parse_lookup(lookup, value, self.choices)
        # if expected is None:
        #    assert res is None
        #    assert re.fullmatch(error_re, err)
        # else:
        assert expected is not None
        assert err is None
        assert res == expected

    def check_format(self, value, expected):
        assert self.type_.get_formatter(self.choices)(value) == expected


class TestQuery:
    def test_from_request(self, query):
        q = Query.from_request("app.model", "fa+1,fd-0,fn", [("bob__equals", "fred")])
        assert q == query

    def test_from_request_duplicate_field(self, query):
        q = Query.from_request(
            "app.model", "fa+1,fd-0,fn,&fa-2", [("bob__equals", "fred")]
        )
        assert q == query

    def test_from_request_with_limit(self, query):
        q = Query.from_request(
            "app.model", "fa+1,fd-0,fn", [("limit", "123"), ("bob__equals", "fred")]
        )
        assert q.arguments["limit"] == 123

    def test_from_request_with_bad_limit(self, query):
        q = Query.from_request(
            "app.model", "fa+1,fd-0,fn", [("limit", "bob"), ("bob__equals", "fred")]
        )
        assert q.arguments["limit"] == 1000

    def test_from_request_with_low_limit(self, query):
        q = Query.from_request(
            "app.model", "fa+1,fd-0,fn", [("limit", "0"), ("bob__equals", "fred")]
        )
        assert q.arguments["limit"] == 1

    def test_from_request_with_related_filter(self):
        q = Query.from_request("app.model", "", [("bob__jones__equals", "fred")])
        assert q == Query(
            "app.model", [], [QueryFilter("bob__jones", "equals", "fred")]
        )

    def test_from_request_with_missing(self):
        q = Query.from_request("app.model", ",,", [])
        assert q == Query("app.model", [], [])

    def test_from_request_filter_no_value(self):
        q = Query.from_request("app.model", "", [("joe__equals", "")])
        assert q == Query("app.model", [], [QueryFilter("joe", "equals", "")])

    def test_from_request_filter_no_lookup(self):
        q = Query.from_request("app.model", "", [("joe", "tom")])
        assert q == Query("app.model", [], [], {"joe": "tom"})

    def test_from_request_filter_bad_lookup(self):
        q = Query.from_request("app.model", "", [("joe__blah", "123")])
        assert q == Query("app.model", [], [QueryFilter("joe", "blah", "123")])

    def test_from_request_filter_no_name(self):
        q = Query.from_request("app.model", "", [("", "123")])
        assert q == Query("app.model", [], [], {"": "123"})

    def test_from_request_field_no_name(self):
        q = Query.from_request("app.model", "+2", [])
        assert q == Query("app.model", [QueryField("", False, ASC, 2)], [])

    def test_from_request_field_no_priority(self):
        q = Query.from_request("app.model", "fn+", [])
        assert q == Query("app.model", [QueryField("fn")], [])

    def test_from_request_field_bad_priority(self):
        q = Query.from_request("app.model", "fn+x", [])
        assert q == Query("app.model", [QueryField("fn")], [])

    def test_from_request_field_pivoted(self):
        q = Query.from_request("app.model", "&fn", [])
        assert q == Query("app.model", [QueryField("fn", True)], [])

    def test_url(self, query):
        query.arguments = {"limit": "123"}
        assert (
            query.get_url("html")
            == "/query/app.model/fa+1,fd-0,fn.html?bob__equals=fred&limit=123"
        )
        expected = (
            "/data_browser/query/app.model/fa+1,fd-0,fn.html?bob__equals=fred&limit=123"
        )
        assert query.get_full_url("html") == expected

    def test_url_no_filters(self, query):
        query.filters = []
        assert query.get_url("html") == "/query/app.model/fa+1,fd-0,fn.html?limit=1000"
        assert (
            query.get_full_url("html")
            == "/data_browser/query/app.model/fa+1,fd-0,fn.html?limit=1000"
        )


class TestBoundQuery:
    def test_fields(self, bound_query):
        assert [f.path for f in bound_query.fields] == [["fa"], ["fd"], ["fn"]]

    def test_sort_fields(self, bound_query):
        assert [(f.path, f.direction, f.priority) for f in bound_query.sort_fields] == [
            (["fd"], DSC, 0),
            (["fa"], ASC, 1),
            (["fn"], None, None),
        ]

    def test_bad_field(self, orm_models):
        query = Query("app.model", [QueryField("yata")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == []

    def test_bad_field_lookup(self, orm_models):
        query = Query("app.model", [QueryField("fa__count__bob")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == []

    def test_bad_fk(self, orm_models):
        query = Query("app.model", [QueryField("yata__yata")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == []

    def test_bad_fk_field(self, orm_models):
        query = Query("app.model", [QueryField("tom__yata")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == []

    def test_bad_fk_field_aggregate(self, orm_models):
        query = Query("app.model", [QueryField("tom__jones__yata")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == []

    def test_bad_long_fk(self, orm_models):
        query = Query("app.model", [QueryField("yata__yata__yata")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == []

    def test_aggregate(self, orm_models):
        query = Query("app.model", [QueryField("tom__jones__count")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == [["tom", "jones", "count"]]

    def test_piovt_aggregate(self, orm_models):
        query = Query("app.model", [QueryField("tom__jones__count", pivoted=True)], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.pivoted for f in bound_query.fields] == [False]

    def test_piovt(self, orm_models):
        query = Query("app.model", [QueryField("tom__jones", pivoted=True)], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.pivoted for f in bound_query.fields] == [True]

    def test_bad_filter(self, orm_models):
        query = Query("app.model", [], [QueryFilter("yata", "equals", "fred")])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.filters] == [["yata"]]

    def test_bad_filter_value(self, orm_models):
        query = Query(
            "app.model",
            [],
            [QueryFilter("num", "equals", "fred"), QueryFilter("num", "equals", 1)],
        )
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.value for f in bound_query.filters] == ["fred", 1]
        assert [f.value for f in bound_query.valid_filters] == [1]

    def test_fk(self, orm_models):
        query = Query("app.model", [QueryField("tom")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == []

    def test_filter_calculated_field(self, orm_models):
        query = Query("app.model", [], [QueryFilter("fn", "equals", "fred")])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.filters] == [["fn"]]


class TestType:
    def test_repr(self):
        assert repr(StringType) == "StringType"


class TestRegexType(ParseHelpers):
    type_ = RegexType

    @pytest.mark.django_db
    @pytest.mark.parametrize(
        "value,expected,err", [(".*", ".*", None), ("\\", None, "Invalid regex")]
    )
    def test_parse(self, value, expected, err):
        self.check_parse(value, expected, err)


class TestStringType(ParseHelpers):
    type_ = StringType

    def test_default_lookup(self):
        assert StringType.default_lookup == "equals"

    @pytest.mark.parametrize("value,expected", [("bob", "bob"), (None, None)])
    def test_format(self, value, expected):
        self.check_format(value, expected)


class TestNumberType(ParseHelpers):
    type_ = NumberType

    @pytest.mark.parametrize(
        "value,expected,err",
        [
            ("6.1", 6.1, None),
            ("hello", None, "Could not convert string to float: 'hello'"),
        ],
    )
    def test_parse(self, value, expected, err):
        self.check_parse(value, expected, err)

    def test_default_lookup(self):
        assert self.type_.default_lookup == "equals"

    @pytest.mark.parametrize("value,expected", [(6, 6), (None, None)])
    def test_format(self, value, expected):
        self.check_format(value, expected)


def dt(Y, M, D, h=0, m=0, s=0):
    return datetime(Y, M, D, h, m, s, tzinfo=timezone.utc)


class TestDateTimeType(ParseHelpers):
    type_ = DateTimeType

    @time_machine.travel("2020-12-13 09:42:53", tick=False)
    @pytest.mark.parametrize(
        "value,expected,err",
        [
            ("2018-03-20T22:31:23", dt(2018, 3, 20, 22, 31, 23), None),
            ("2018-3-2T22:31:23", dt(2018, 3, 2, 22, 31, 23), None),
            ("2018-3-2 22:31:23", dt(2018, 3, 2, 22, 31, 23), None),
            ("2018-3-2 days+1 22:31:23", dt(2018, 3, 3, 22, 31, 23), None),
            ("2018-3-2 days+1 2:31", dt(2018, 3, 3, 2, 31, 0), None),
            ("hello", None, "Unrecognized clause 'hello'"),
            ("now", dt(2020, 12, 13, 9, 42, 53), None),
            ("now hours+1", dt(2020, 12, 13, 10, 42, 53), None),
            ("today", dt(2020, 12, 13, 0, 0, 0), None),
            # dateutil.parser
            ("11-11-2018", dt(2018, 11, 11), None),
            ("11-22-2018", dt(2018, 11, 22), None),
            ("21-12-2018", dt(2018, 12, 21), None),
            ("11-12-2018", None, "Ambiguous value"),
            ("21-22-2018", None, "Unrecognized clause '21-22-2018'"),
            ("2018-11-11", dt(2018, 11, 11), None),
            ("2018-11-22", dt(2018, 11, 22), None),
            ("2018-21-12", None, "Month must be in 1..12: 2018-21-12"),
            ("2018-11-12", dt(2018, 11, 12), None),
            ("2018-21-22", None, "Month must be in 1..12: 2018-21-22"),
            ("2018-3-2", dt(2018, 3, 2), None),
            ("20181111", dt(2018, 11, 11), None),
            ("20181122", dt(2018, 11, 22), None),
            ("20182112", None, "Month must be in 1..12: 20182112"),
            ("20181112", dt(2018, 11, 12), None),
            ("20182122", None, "Month must be in 1..12: 20182122"),
            ("181111", None, "Ambiguous value"),
            ("181122", None, "Ambiguous value"),
            ("182112", None, "Unrecognized clause '182112'"),
            ("181112", None, "Ambiguous value"),
            ("182122", None, "Unrecognized clause '182122'"),
            ("111118", None, "Ambiguous value"),
            ("112218", dt(2018, 11, 22), None),
            ("211218", None, "Ambiguous value"),
            ("111218", None, "Ambiguous value"),
            ("212218", None, "Unrecognized clause '212218'"),
            # dateutil.relativedelta
            ("hour+1", dt(2020, 12, 13, 10, 42, 53), None),
            ("hour-1", dt(2020, 12, 13, 8, 42, 53), None),
            ("hour=1", dt(2020, 12, 13, 1, 42, 53), None),
            ("hour 1", None, "Unrecognized clause 'hour'"),
            ("hour+0", dt(2020, 12, 13, 9, 42, 53), None),
            ("hour=0", dt(2020, 12, 13, 0, 42, 53), None),
            ("month+1", dt(2021, 1, 13, 9, 42, 53), None),
            ("month-1", dt(2020, 11, 13, 9, 42, 53), None),
            ("month=1", dt(2020, 1, 13, 9, 42, 53), None),
            ("month 1", None, "Unrecognized clause 'month'"),
            ("month+0", dt(2020, 12, 13, 9, 42, 53), None),
            ("month=0", None, "Can't set 'month' to '0'"),
            ("thurs+1", dt(2020, 12, 17, 9, 42, 53), None),
            ("thurs-1", dt(2020, 12, 10, 9, 42, 53), None),
            ("thurs=1", None, "'=' not supported for 'thurs'"),
            ("thurs 1", None, "Unrecognized clause 'thurs'"),
            ("thurs+0", None, "Can't add '0' 'thurs's"),
            ("thurs=0", None, "'=' not supported for 'thurs'"),
            ("month+1 month=1", dt(2021, 1, 13, 9, 42, 53), None),
            ("month=1 month+1", dt(2020, 2, 13, 9, 42, 53), None),
            ("bobit+1", None, "Unrecognized field 'bobit'"),
            ("week=1", None, "'=' not supported for 'week'"),
        ],
    )
    def test_parse(self, value, expected, err):
        self.check_parse(value, expected, err)

    def test_default_lookup(self):
        assert DateTimeType.default_lookup == "equals"

    @pytest.mark.django_db
    @pytest.mark.parametrize(
        "value,expected",
        [
            (
                datetime(2020, 5, 19, 8, 42, 16, tzinfo=timezone.utc),
                "2020-05-19 08:42:16",
            ),
            (None, None),
        ],
    )
    def test_aware_format(self, value, expected, settings):
        settings.USE_TZ = True
        self.check_format(value, expected)

    @pytest.mark.django_db
    @pytest.mark.parametrize(
        "value,expected",
        [(datetime(2020, 5, 19, 8, 42, 16), "2020-05-19 08:42:16"), (None, None)],
    )
    def test_naive_format(self, value, expected, settings):
        settings.USE_TZ = False
        self.check_format(value, expected)


class TestDurationType(ParseHelpers):
    type_ = DurationType

    @pytest.mark.parametrize(
        "value,expected,err",
        [
            ("5 days", timedelta(days=5), None),
            ("5 5:5", timedelta(days=5, hours=5, minutes=5), None),
            ("5 dayss", None, "Duration value should be 'DD HH:MM:SS'"),
        ],
    )
    def test_parse(self, value, expected, err):
        self.check_parse(value, expected, err)

    def test_default_lookup(self):
        assert self.type_.default_lookup == "equals"

    @pytest.mark.parametrize(
        "value,expected",
        [
            (timedelta(days=5, minutes=6), "5 days, 0:06:00"),
            (timedelta(minutes=6), "0:06:00"),
            (None, None),
        ],
    )
    def test_format(self, value, expected):
        self.check_format(value, expected)


class TestDateType(ParseHelpers):
    type_ = DateType

    @time_machine.travel("2020-12-13 09:42:53", tick=False)
    @pytest.mark.parametrize(
        "value,expected,err",
        [
            ("2018-03-20T22:31:23", date(2018, 3, 20), None),
            ("hello", None, "Unrecognized clause 'hello'"),
            ("today", date(2020, 12, 13), None),
            ("today days+1", date(2020, 12, 14), None),
            ("now", date(2020, 12, 13), None),
            ("11-22-2018", date(2018, 11, 22), None),
            ("21-12-2018", date(2018, 12, 21), None),
            ("11-12-2018", None, "Ambiguous value"),
            ("21-22-2018", None, "Unrecognized clause '21-22-2018'"),
            ("days+1", date(2020, 12, 14), None),
        ],
    )
    def test_parse(self, value, expected, err):
        self.check_parse(value, expected, err)

    def test_default_lookup(self):
        assert self.type_.default_lookup == "equals"

    @pytest.mark.parametrize(
        "value,expected", [(date(2020, 5, 19), "2020-05-19"), (None, None)]
    )
    def test_format(self, value, expected):
        self.check_format(value, expected)


class TestBooleanType(ParseHelpers):
    type_ = BooleanType

    @pytest.mark.parametrize(
        "value,expected,err",
        [
            ("True", True, None),
            ("False", False, None),
            ("hello", None, "Expected 'true' or 'false'"),
        ],
    )
    def test_parse(self, value, expected, err):
        self.check_parse(value, expected, err)

    @pytest.mark.parametrize("value,expected", [(None, None)])
    def test_format(self, value, expected):
        self.check_format(value, expected)

    def test_default_lookup(self):
        assert self.type_.default_lookup == "equals"


class TestStringChoiceType(ParseHelpers):
    type_ = StringChoiceType
    choices = [("a", "A"), ("b", "B"), ("c", "C")]

    @pytest.mark.parametrize("value,expected", [("b", "B"), (None, None)])
    def test_format(self, value, expected):
        self.check_format(value, expected)

    def test_bad_value(self):
        assert self.type_.get_formatter(self.choices)("x") == "x"

    @pytest.mark.parametrize(
        "value,expected,err", [("B", "b", None), ("X", None, "Unknown choice 'X'")]
    )
    def test_parse(self, value, expected, err):
        self.check_parse(value, expected, err)


class TestNumberChoiceType(ParseHelpers):
    type_ = NumberChoiceType
    choices = [(1, "A"), (2, "B"), (3, "C")]

    @pytest.mark.parametrize("value,expected", [(2, "B"), (None, None)])
    def test_format(self, value, expected):
        self.check_format(value, expected)

    def test_bad_value(self):
        assert self.type_.get_formatter(self.choices)(6) == 6

    @pytest.mark.parametrize(
        "value,expected,err", [("B", 2, None), ("X", None, "Unknown choice 'X'")]
    )
    def test_parse(self, value, expected, err):
        self.check_parse(value, expected, err)


class TestHTMLType(ParseHelpers):
    type_ = HTMLType

    @pytest.mark.parametrize("value,expected", [(None, None)])
    def test_format(self, value, expected):
        self.check_format(value, expected)


class TestIsNullType(ParseHelpers):
    type_ = IsNullType

    @pytest.mark.parametrize(
        "value,expected", [(True, "IsNull"), (False, "NotNull"), (None, "IsNull")]
    )
    def test_format(self, value, expected):
        self.check_format(value, expected)

    @pytest.mark.parametrize(
        "value,expected,err", [("IsNull", True, None), ("NotNull", False, None)]
    )
    def test_parse(self, value, expected, err):
        self.check_parse(value, expected, err)


class TestUUIDType(ParseHelpers):
    type_ = UUIDType
    s = "210081a4-7315-4173-aa73-4bab47b2f9a5"
    u = UUID(s)

    @pytest.mark.parametrize("value,expected", [(u, s), (None, None)])
    def test_format(self, value, expected):
        self.check_format(value, expected)

    @pytest.mark.parametrize(
        "value,expected,err",
        [
            (s, u, None),
            (s.replace("-", ""), u, None),
            (s[1:], None, "Badly formed hexadecimal UUID string"),
        ],
    )
    def test_parse(self, value, expected, err):
        self.check_parse(value, expected, err)


class TestUnknownType(ParseHelpers):
    type_ = UnknownType

    @pytest.mark.parametrize("value,expected", [(None, None)])
    def test_format(self, value, expected):
        self.check_format(value, expected)


class TestStringArrayType(ParseHelpers):
    type_ = StringArrayType

    @pytest.mark.parametrize(
        "value,expected,err",
        [('["bob"]', ["bob"], None), ('"bob"', None, "Expected a list")],
    )
    def test_parse(self, value, expected, err):
        self.check_parse(value, expected, err)


class TestStringChoiceArrayType(ParseHelpers):
    type_ = StringChoiceArrayType
    choices = [("fred", "bob")]

    @pytest.mark.parametrize(
        "value,expected,err",
        [('["bob"]', ["fred"], None), ('"bob"', None, "Expected a list")],
    )
    def test_parse(self, value, expected, err):
        self.check_parse(value, expected, err)

    @pytest.mark.parametrize(
        "lookup,value,expected,err",
        [("contains", "bob", "fred", None), ("length", "1", 1, None)],
    )
    def test_parse_lookup(self, lookup, value, expected, err):
        self.check_parse_lookup(lookup, value, expected, err)


class TestNumberArrayType(ParseHelpers):
    type_ = NumberArrayType

    @pytest.mark.parametrize(
        "value,expected,err", [("[123]", [123], None), ("123", None, "Expected a list")]
    )
    def test_parse(self, value, expected, err):
        self.check_parse(value, expected, err)


class TestNumberChoiceArrayType(ParseHelpers):
    type_ = NumberChoiceArrayType
    choices = [(123, "bob")]

    @pytest.mark.parametrize(
        "value,expected,err",
        [('["bob"]', [123], None), ('"bob"', None, "Expected a list")],
    )
    def test_parse(self, value, expected, err):
        self.check_parse(value, expected, err)

    @pytest.mark.parametrize(
        "lookup,value,expected,err",
        [("contains", "bob", 123, None), ("length", "1", 1, None)],
    )
    def test_parse_lookup(self, lookup, value, expected, err):
        self.check_parse_lookup(lookup, value, expected, err)



import dataclasses
import urllib
from typing import Any, Mapping, Optional, Sequence

from django.urls import reverse
from django.utils.functional import cached_property

from .common import settings
from .types import ASC, DSC


@dataclasses.dataclass
class QueryField:
    path: str
    pivoted: bool = False
    direction: str = None
    priority: int = None

    def __post_init__(self):
        self.path = self.path.split("__")
        assert (self.direction is None) == (self.priority is None)


@dataclasses.dataclass
class QueryFilter:
    path: str
    lookup: str
    value: str

    def __post_init__(self):
        self.path_str = self.path
        self.path = self.path.split("__")


def parse_sort(value, symbol, direction):
    path, priority = value.split(symbol)
    try:
        return path, direction, int(priority)
    except:  # noqa: E722  input sanitization
        return path, None, None


@dataclasses.dataclass
class Query:
    model_name: str
    fields: Sequence[QueryField]
    filters: Sequence[QueryFilter]
    arguments: Mapping[str, str] = dataclasses.field(default_factory=dict)

    @classmethod
    def from_request(cls, model_name, field_str, params):
        fields = []
        found_fields = set()
        for field in field_str.split(","):
            field = field.strip()
            if field:
                # pivot
                if field.startswith("&"):
                    field = field[1:]
                    pivoted = True
                else:
                    pivoted = False

                # sort
                if "+" in field:
                    path, direction, priority = parse_sort(field, "+", ASC)
                elif "-" in field:
                    path, direction, priority = parse_sort(field, "-", DSC)
                else:
                    path, direction, priority = field, None, None

                # collect
                if path not in found_fields:
                    found_fields.add(path)
                    fields.append(QueryField(path, pivoted, direction, priority))

        filters = []
        arguments = {}
        for path__lookup, value in params:
            if "__" in path__lookup:
                path, lookup = path__lookup.rsplit("__", 1)
                filters.append(QueryFilter(path, lookup, value))
            else:
                arguments[path__lookup] = value

        return cls(model_name, fields, filters, arguments)

    def __post_init__(self):
        # sanitize limit
        value = self.arguments.get("limit")
        try:
            value = int(value)
        except:  # noqa: E722  input sanitization
            value = settings.DATA_BROWSER_DEFAULT_ROW_LIMIT
        self.arguments["limit"] = max(1, value)

    @property
    def _field_str(self):
        field_strs = []
        for field in self.fields:
            pivot = "&" if field.pivoted else ""
            direction = {ASC: "+", DSC: "-", None: ""}[field.direction]
            priority = str(field.priority) if field.direction else ""
            field_strs.append(f"{pivot}{'__'.join(field.path)}{direction}{priority}")
        return ",".join(field_strs)

    @property
    def _filter_fields(self):
        return [
            ("__".join(filter.path + [filter.lookup]), filter.value)
            for filter in self.filters
        ] + [(key, str(value)) for key, value in self.arguments.items()]

    def get_url(self, media):
        url = self.get_full_url(media)
        home = reverse("data_browser:home")
        assert url.startswith(home)
        return url[len(home) - 1 :]

    def get_full_url(self, media):
        base_url = reverse(
            "data_browser:query",
            kwargs={
                "model_name": self.model_name,
                "fields": self._field_str,
                "media": media,
            },
        )
        params = urllib.parse.urlencode(
            self._filter_fields, quote_via=urllib.parse.quote_plus, doseq=True
        )
        return f"{base_url}?{params}"


@dataclasses.dataclass
class BoundFilter:
    orm_bound_field: Any
    lookup: str
    value: str
    query_filter: QueryFilter

    @classmethod
    def bind(cls, orm_bound_field, query_filter):
        return cls(
            orm_bound_field, query_filter.lookup, query_filter.value, query_filter
        )

    def __post_init__(self):
        if self.orm_bound_field:
            self.parsed, self.err_message = self.orm_bound_field.parse_lookup(
                self.lookup, self.value
            )
        else:
            self.err_message = "Unknown field"
        self.is_valid = not self.err_message

    @cached_property
    def path(self):
        return self.query_filter.path

    @cached_property
    def path_str(self):
        return self.query_filter.path_str

    def formatted_value(self):
        if not self.is_valid:
            return None
        return self.orm_bound_field.format_lookup(self.lookup, self.parsed)


@dataclasses.dataclass
class BoundField:
    orm_bound_field: Any
    pivoted: bool
    direction: Optional[str]
    priority: Optional[int]

    @classmethod
    def bind(cls, orm_bound_field, query_field):
        return cls(
            orm_bound_field,
            query_field.pivoted and orm_bound_field.can_pivot,
            query_field.direction if orm_bound_field.concrete else None,
            query_field.priority if orm_bound_field.concrete else None,
        )

    @cached_property
    def path(self):
        return self.orm_bound_field.full_path

    @cached_property
    def pretty_path(self):
        return self.orm_bound_field.pretty_path

    @cached_property
    def path_str(self):
        return self.orm_bound_field.path_str


def _orm_fields(fields):
    return [f.orm_bound_field for f in fields]


class BoundQuery:
    def __init__(self, model_name, fields, filters, limit, orm_model):
        self.model_name = model_name
        self.fields = fields
        self.filters = filters
        self.limit = limit
        self.orm_model = orm_model

    @classmethod
    def bind(cls, query, orm_models):
        def get_orm_field(parts):
            model_name = query.model_name
            orm_bound_field = None
            for part in parts:
                if model_name is None:
                    return None
                orm_field = orm_models[model_name].fields.get(part)
                if orm_field is None:
                    return None
                orm_bound_field = orm_field.bind(orm_bound_field)
                model_name = orm_field.rel_name
            if not orm_bound_field.type_:
                return None
            return orm_bound_field

        model_name = query.model_name

        fields = []
        for query_field in query.fields:
            orm_bound_field = get_orm_field(query_field.path)
            if orm_bound_field:
                fields.append(BoundField.bind(orm_bound_field, query_field))

        filters = []
        for query_filter in query.filters:
            orm_bound_field = get_orm_field(query_filter.path)
            if orm_bound_field and not orm_bound_field.concrete:
                orm_bound_field = None
            filters.append(BoundFilter.bind(orm_bound_field, query_filter))

        return cls(
            model_name=model_name,
            fields=fields,
            filters=filters,
            limit=query.arguments["limit"],
            orm_model=orm_models[model_name],
        )

    @cached_property
    def sort_fields(self):
        res = sorted((f for f in self.fields if f.direction), key=lambda f: f.priority)
        # add all unsorted fields on the end of the sort to make stuff stable
        res += sorted((f for f in self.fields if not f.direction), key=lambda f: f.path)
        return res

    @cached_property
    def valid_filters(self):
        return [f for f in self.filters if f.is_valid]

    @cached_property
    def col_fields(self):
        return [f for f in self.fields if f.pivoted]

    @cached_property
    def row_fields(self):
        return [f for f in self.fields if f.orm_bound_field.can_pivot and not f.pivoted]

    @cached_property
    def body_fields(self):
        return [f for f in self.fields if not f.orm_bound_field.can_pivot]

    @cached_property
    def bound_fields(self):
        return _orm_fields(self.fields)

    @cached_property
    def bound_filters(self):
        return _orm_fields(self.valid_filters)

    @cached_property
    def bound_col_fields(self):
        return _orm_fields(self.col_fields)

    @cached_property
    def bound_row_fields(self):
        return _orm_fields(self.row_fields)

    @cached_property
    def bound_body_fields(self):
        return _orm_fields(self.body_fields)



class ANY:
    def __init__(self, type):
        self.type = type

    def __eq__(self, other):
        return isinstance(other, self.type)


class KEYS:
    def __init__(self, *keys):
        self.keys = set(keys)

    def __eq__(self, other):
        return isinstance(other, dict) and other.keys() == self.keys



from django.apps import AppConfig
from django.db.models import AutoField


class DataBrowserConfig(AppConfig):
    name = "data_browser"
    verbose_name = "Data Browser"
    default_auto_field = AutoField
